@@ -36,7 +36,7 @@ import datetime
 from traceback import format_tb
 
 
-# from contextlib import contextmanager
+from contextlib import contextmanager
 
 import h5py as h5
 import numpy as np
@@ -48,7 +48,15 @@ from datamate.namespaces import Namespace, namespacify
 
 __all__ = ["Directory", "ArrayFile"]
 
-# -- Custom Errors --------------------------------------------------------------
+# -- Custom Errors and Warnings ------------------------------------------------
+
+
+class ConfigWarning(Warning):
+    pass
+
+
+class ModifiedWarning(Warning):
+    pass
 
 
 class ModifiedError(Exception):
@@ -102,6 +110,9 @@ def root(root_dir: Union[str, Path, NoneType] = None):
 
     root_dir (str or Path): root directory that will be set at execution of the
         callable. Optional. Default is None which corresponds to get_root_dir().
+
+    Note, this will take precedence over set_root_dir, and set_root_context context when
+    the callable is called within the context.
     """
 
     def decorator(callable):
@@ -118,6 +129,20 @@ def root(root_dir: Union[str, Path, NoneType] = None):
     return decorator
 
 
+@contextmanager
+def set_root_context(root_dir: Union[str, Path, NoneType] = None):
+    """Set root directory within a context and revert after.
+
+    Example:
+        with set_root_context(dir):
+            Directory(config)
+    """
+    _root_dir = get_root_dir()
+    set_root_dir(root_dir)
+    yield
+    set_root_dir(_root_dir)
+
+
 def enforce_config_match(enforce: bool) -> None:
     """
     Enforce error if configs are not matching.
@@ -176,7 +201,7 @@ def get_scope() -> Dict[str, type]:
     return context.scope if hasattr(context, "scope") else get_default_scope()
 
 
-def get_default_scope() -> Dict[str, type]:
+def get_default_scope(cls: object = None) -> Dict[str, type]:
     """
     Return the default scope used for "type" field resolution.
     """
@@ -184,21 +209,31 @@ def get_default_scope() -> Dict[str, type]:
     def subclasses(t: type) -> Iterator[type]:
         yield from itertools.chain([t], *map(subclasses, t.__subclasses__()))
 
+    cls = cls or Directory
     scope: Dict[str, type] = {}
-    for t in subclasses(Directory):
+    for t in subclasses(cls):
         scope[t.__qualname__] = t
     return scope
 
 
+def reset_scope(cls: object = None) -> None:
+    """
+    Reset the scope to the default scope.
+    """
+    set_scope(get_default_scope(cls))
+
+
 # -- Directorys -----------------------------------------------------------------
 
 
-class NoInit(type):
+class NonExistingDirectory(type):
+    """Directory metaclass to allow create non-existing Directory instances."""
+
     def __call__(cls, *args, **kwargs):
         return cls.__new__(cls, *args, **kwargs)
 
 
-class Directory(metaclass=NoInit):
+class Directory(metaclass=NonExistingDirectory):
     """
     An array- and metadata-friendly view into a directory
 
@@ -222,7 +257,7 @@ class Directory(metaclass=NoInit):
     directory.
 
     If only `conf` is provided, it will search the current `root_dir`
-    for a matching directory, and return an Directory pointing there if it
+    for a matching directory, and return a Directory pointing there if it
     exists. Otherwise, a new Directory will be constructed at the top level of
     the `root_dir`.
 
@@ -414,7 +449,7 @@ class Directory(metaclass=NoInit):
     def parent(self):
         return Directory(self.path.absolute().parent)
 
-    def _override_config(selfconfigec, status=None):
+    def _override_config(self, config, status=None):
         """Overriding config stored in _meta.yaml.
 
         config (Dict): update for meta.config
@@ -435,7 +470,7 @@ class Directory(metaclass=NoInit):
                         f"Overriding config. Diff is:"
                         f'{config.diff(current_config, name1="passed", name2="stored")}'
                     ),
-                    Warning,
+                    ConfigWarning,
                     stacklevel=2,
                 )
             write_meta(config=config, status="overridden")
@@ -454,7 +489,7 @@ class Directory(metaclass=NoInit):
                 warnings.simplefilter("always")
                 warnings.warn(
                     (f"Overriding status {current_status} to {status}"),
-                    Warning,
+                    ConfigWarning,
                     stacklevel=2,
                 )
         write_meta(config=self.config, status=status)
@@ -541,14 +576,9 @@ class Directory(metaclass=NoInit):
         `key` are transformed into ".", to support accessing encoded files as
         attributes (i.e. `Directory['name.ext'] = val` is equivalent to
         `Directory.name__ext = val`).
-
-        TODO: contextmanager that controls if setattr stores in memory or on disk.
         """
         path = self.path / key
 
-        if self.config is not None and self.status == "done":
-            self._modified_past_init(True)
-
         # Copy an existing file or directory.
         if isinstance(val, Path):
             if os.path.isfile(val):
@@ -570,9 +600,16 @@ class Directory(metaclass=NoInit):
                 raise TypeError(
                     format_tb(err.__traceback__)[0]
                     + err.args[0]
-                    + f"\nYou're trying to store {val} which cannot be converted to h5-file."
+                    + f"\nYou're trying to store {val} which cannot be converted to h5-file in {path}."
+                    + "\nFor reference of supported types, see https://docs.h5py.org/en/stable/faq.html?highlight=types#numpy-object-types"
+                    + "\nE.g. NumPy unicode strings must be converted to 'S' strings and back:"
+                    + "\nfoo.bar = array.astype('S') to store and foo.bar[:].astype('U') to retrieve."
                 ) from None
 
+        if self.config is not None and self.status == "done":
+            # Track if a Directory has been modified past __init__
+            self._modified_past_init(True)
+
     def __delitem__(self, key: str) -> None:
         """
         Deletes the entry at `self.path/key`
@@ -592,7 +629,7 @@ class Directory(metaclass=NoInit):
         elif path.is_file():
             path.unlink()
 
-        # Delete an Directory.
+        # Delete a Directory.
         else:
             shutil.rmtree(path, ignore_errors=True)
 
@@ -680,6 +717,10 @@ class Directory(metaclass=NoInit):
         return self.__getitem__(key.replace("__", "."))
 
     def __setattr__(self, key: str, value: object) -> None:
+        # Fix autoreload related effect.
+        if key.startswith("__") and key.endswith("__"):
+            object.__setattr__(self, key, value)
+            return
         self.__setitem__(key.replace("__", "."), value)
 
     def __delattr__(self, key: str) -> None:
@@ -814,7 +855,7 @@ def _check_for_init(cls: type) -> bool:
 
 def _directory_from_path(cls: type, path: Path) -> Directory:
     """
-    Return an Directory corresponding to the file tree at `path`.
+    Return a Directory corresponding to the file tree at `path`.
 
     An error is raised if the type recorded in `_meta.yaml`, if any, is not a
     subtype of `cls`.
@@ -846,27 +887,18 @@ def _directory_from_path(cls: type, path: Path) -> Directory:
 
 def _directory_from_config(cls: type, conf: Mapping[str, object]) -> Directory:
     """
-    Find or build an Directory with the given type and Namespace.
+    Find or build a Directory with the given type and Namespace.
     """
     directory = _forward_subclass(cls, conf)
+    new_dir_path = _new_directory_path(type(directory))
     object.__setattr__(directory, "_cached_keys", set())
     config = Namespace(**directory._config)
 
-    for path in Path(get_root_dir()).glob("*"):
-        meta = read_meta(path)
-
-        if getattr(meta, "modified", False):
-            raise ModifiedError()
-
-        if meta.config == config:
-            while meta.status == "running":
-                sleep(0.01)
-                meta = read_meta(path)
-            if meta.status == "done":
-                object.__setattr__(directory, "path", path)
-                return directory
-    else:
-        object.__setattr__(directory, "path", _new_directory_path(type(directory)))
+    def _new_directory():
+        object.__setattr__(directory, "path", new_dir_path)
+        # return empty path cause only the type field is populated
+        if list(config.keys()) == ["type"]:
+            return directory
         # catches FileExistsError for the case when two processes try to
         # build the same directory simultaneously
         try:
@@ -875,12 +907,48 @@ def _directory_from_config(cls: type, conf: Mapping[str, object]) -> Directory:
             return _directory_from_config(cls, conf)
         return directory
 
+    for path in Path(get_root_dir()).glob("*"):
+        meta = read_meta(path)
+
+        if meta.config == config:
+            if getattr(meta, "modified", False):
+                with warnings.catch_warnings():
+                    warnings.simplefilter("always")
+                    warnings.warn(
+                        (
+                            f"The Directory {path} has been modified after being build."
+                            + f"\nBuilding a new directory {new_dir_path} to prevent that the files you would get from the object"
+                            + " mismatch the files you would expect based on your __init__ and configuration."
+                            + "\nYou can circumvent this"
+                            + " by e.g. using the explicit path as constructor (see Directory docs)."
+                        ),
+                        ModifiedWarning,
+                        stacklevel=2,
+                    )
+                return _new_directory()
+            # raise ModifiedError(
+            #             f"The Directory {path} has been modified after being build."
+            #             + "\nThis could mean that the files you would get from the object"
+            #             + " mismatch the files you would expect based on your configuration."
+            #             + "\nYou can circumvent this error"
+            #             + " by e.g. using the path as constructor"
+            #             + " as explained in the Directory docs."
+            #         )
+
+            while meta.status == "running":
+                sleep(0.01)
+                meta = read_meta(path)
+            if meta.status == "done":
+                object.__setattr__(directory, "path", path)
+                return directory
+    return _new_directory()
+
 
 def _directory_from_path_and_config(
     cls: type, path: Path, conf: Mapping[str, object]
 ) -> Directory:
     """
-    Find or build an Directory with the given type, path, and Namespace.
+    Find or build a Directory with the given type, path, and Namespace.
     """
     directory = _forward_subclass(cls, conf)
     object.__setattr__(directory, "_cached_keys", set())
@@ -1026,8 +1094,15 @@ def _forward_subclass(cls: type, config: object = {}) -> object:
     elif isinstance(cls_override, str):
         try:
             cls = get_scope()[cls_override]
-        except:
-            raise KeyError(f'"{cls_override}" can\'t be resolved.')
+        except KeyError as e:
+            raise KeyError(
+                f'"{cls_override}" can\'t be resolved because it is not found'
+                + f" inside the scope of Directory subclasses. Typo?"
+                + " If this happens in the context of autoreload enabled in"
+                + " a notebook/IPython session, run `datamate.reset_scope(datamate.Directory)`"
+                + " as a workaround or restart the kernel"
+                + f" (background: https://github.com/ipython/ipython/issues/12399)."
+            ) from e
 
     # Construct and return a `Configurable` instance.
     obj = object.__new__(cls)
@@ -1153,8 +1228,9 @@ def read_meta(path: Path) -> Namespace:
         assert isinstance(meta, Namespace)
         if hasattr(meta, "config"):
             assert isinstance(meta.config, Namespace)
-        elif hasattr(meta, "spec"):
+        elif hasattr(meta, "spec"):  # for backwards compatibility
             assert isinstance(meta.spec, Namespace)
+            meta["config"] = meta.pop("spec")
         assert isinstance(meta.status, str)
         return meta
     except:
@@ -1174,7 +1250,7 @@ def directory_to_df(directory: Directory, dtypes: dict = None) -> DataFrame:
     def convert(_arr):
         if isinstance(_arr, np.ndarray):
             if isinstance(_arr.item(0), (np.character, bytes)):
-                return _arr.astype(np.str)
+                return _arr.astype(str)
         return _arr
 
     df_dict = {
@@ -1197,11 +1273,7 @@ def directory_to_df(directory: Directory, dtypes: dict = None) -> DataFrame:
 
     df_dict = valmap(
         convert,
-        {
-            key: val.tolist()
-            for key, val in df_dict.items()
-            if len(val) == most_frequent_length
-        },
+        {key: val for key, val in df_dict.items() if len(val) == most_frequent_length},
     )
     if dtypes is not None:
         df_dict = {
